@@ -77,8 +77,10 @@ typedef enum FunctionCode{
   F_ENDFUNC,
   F_ENDPOPUP,
   F_TAKE_SCREENSHOT,
+  F_SET,
   F_Test,    /* for debugging purposes to be able to test new features before actually
               * enabling them for user */
+			 
   /* this functions require window as aparameter */
   F_WINDOW_FUNC_START,
   F_MOVE,
@@ -228,8 +230,10 @@ typedef struct MenuData
     struct MenuDataItem *last;  /* last item in menu */
     short items_num;        /* number of items in the menu */
 
-	unsigned int    recent_items ;
+	int    recent_items ;   /*  negative value indicates that default, 
+								set in feel should be used */
 
+	char  *comment ;
 }MenuData;
 
 #define MAX_MENU_ITEM_HEIGHT    (ASDefaultScrHeight>>4)
@@ -294,8 +298,10 @@ MenuData *find_menu_data( ASHashTable *list, char *name );
 MenuDataItem *new_menu_data_item( MenuData *menu );
 int parse_menu_item_name (MenuDataItem * item, char **name);
 
+Bool check_fdata_availability( FunctionData *fdata );
+
 MenuDataItem * add_menu_fdata_item( MenuData *menu, FunctionData *fdata, char *minipixmap, struct ASImage *img );
-MenuDataItem * menu_data_item_from_func (MenuData * menu, FunctionData * fdata);
+MenuDataItem * menu_data_item_from_func (MenuData * menu, FunctionData * fdata, Bool do_check_availability);
 struct ASImage *check_scale_menu_pmap( struct ASImage *im, ASFlagType flags ); 
 void reload_menu_pmaps( MenuData *menu );
 
@@ -305,12 +311,7 @@ int compare_func_data_name(const void *a, const void *b);
 
 void print_func_data(const char *file, const char *func, int line, FunctionData *data);
 
-#if defined(LOCAL_DEBUG) || defined(DEBUG)
-#define MenuDataItemFromFunc(m,f) \
-do{fprintf(stderr,"MenuDataItemFromFunc called:");print_func_data(__FILE__, __FUNCTION__, __LINE__,f);menu_data_item_from_func((m),(f));}while(0)
-#else
-#define MenuDataItemFromFunc(m,f) menu_data_item_from_func((m),(f))
-#endif
+#define MenuDataItemFromFunc(m,f) menu_data_item_from_func((m),(f), False)
 
 
 
